@@ -23,8 +23,10 @@ BASE = os.path.dirname(os.path.abspath(__file__))
 
 def load_shaders(name):
     from .fast_data_types import GLSL_VERSION
-    vert = open(os.path.join(BASE, '{}_vertex.glsl'.format(name))).read().replace('GLSL_VERSION', str(GLSL_VERSION), 1)
-    frag = open(os.path.join(BASE, '{}_fragment.glsl'.format(name))).read().replace('GLSL_VERSION', str(GLSL_VERSION), 1)
+    with open(os.path.join(BASE, '{}_vertex.glsl'.format(name))) as f:
+        vert = f.read().replace('GLSL_VERSION', str(GLSL_VERSION), 1)
+    with open(os.path.join(BASE, '{}_fragment.glsl'.format(name))) as f:
+        frag = f.read().replace('GLSL_VERSION', str(GLSL_VERSION), 1)
     return vert, frag
 
 
@@ -162,7 +164,10 @@ def open_cmd(cmd, arg=None, cwd=None):
     import subprocess
     if arg is not None:
         cmd = list(cmd)
-        cmd.append(arg)
+        if isinstance(arg, (list, tuple)):
+            cmd.extend(arg)
+        else:
+            cmd.append(arg)
     return subprocess.Popen(cmd, stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, cwd=cwd or None)
 
 
@@ -413,15 +418,12 @@ def get_editor():
     ans = getattr(get_editor, 'ans', False)
     if ans is False:
         import shlex
-        ans = os.environ.get('EDITOR')
-        if not ans or not exe_exists(shlex.split(ans)[0]):
-            for q in ('vim', 'nvim', 'vi', 'emacs', 'micro', 'nano', 'vis'):
-                r = exe_exists(q)
-                if r:
-                    ans = r
-                    break
-            else:
-                ans = 'vim'
+        for ans in (os.environ.get('VISUAL'), os.environ.get('EDITOR'), 'vim',
+                    'nvim', 'vi', 'emacs', 'kak', 'micro', 'nano', 'vis'):
+            if ans and exe_exists(shlex.split(ans)[0]):
+                break
+        else:
+            ans = 'vim'
         ans = shlex.split(ans)
         get_editor.ans = ans
     return ans
